import time


class ManagerLost(Exception):
    """
    Task lost due to manager loss.  Manager is considered lost when multiple heartbeats
    have been missed.
    """
    def __init__(self, manager_id: bytes, hostname: str) -> None:
        self.manager_id = manager_id
        self.tstamp = time.time()
        self.hostname = hostname

    def __str__(self) -> str:
        return (
            f"Task failure due to loss of manager {self.manager_id.decode()} on"
            f" host {self.hostname}"
        )


class VersionMismatch(Exception):
    """Manager and Interchange versions do not match"""
    def __init__(self, interchange_version: str, manager_version: str):
        self.interchange_version = interchange_version
        self.manager_version = manager_version

    def __str__(self) -> str:
        return (
            f"Manager version info {self.manager_version} does not match interchange"
            f" version info {self.interchange_version}"
        )


class WorkerLost(Exception):
    """Exception raised when a worker is lost
    """
    def __init__(self, worker_id, hostname):
        self.worker_id = worker_id
        self.hostname = hostname

    def __str__(self):
        return "Task failure due to loss of worker {} on host {}".format(self.worker_id, self.hostname)


class CommandClientTimeoutError(Exception):
    """Raised when the command client times out waiting for a response.
    """


class CommandClientBadError(Exception):
    """Raised when the command client is bad from an earlier timeout.
    """



import pathlib
from typing import Callable

from parsl.logconfigs.base import LogConfig


class NoopLogging(LogConfig):
    """A log configuration which doesn't configure any logging.

    This is intended to help with benchmarking, but can also be used to
    configure an environment with no logging from any LogConfig aware
    component.
    """

    def initialize_logging(self, *, log_dir: pathlib.Path, log_name: str) -> Callable[[], None]:

        def unregister_callback() -> None:
            pass

        return unregister_callback



from unittest import mock

import pytest

from parsl import HighThroughputExecutor
from parsl.executors.high_throughput import zmq_pipes


@pytest.mark.local
def test_submit_payload():
    htex = HighThroughputExecutor()
    htex.outgoing_q = mock.Mock(spec=zmq_pipes.TasksOutgoing)
    ctxt = {"some": "context"}
    buf = b'some buffer (function) payload'
    for task_num in range(1, 20):
        htex.outgoing_q.reset_mock()
        fut = htex.submit_payload(ctxt, buf)
        (msg,), _ = htex.outgoing_q.put.call_args

        assert htex.tasks[fut.parsl_executor_task_id] is fut
        assert fut.parsl_executor_task_id == task_num, "Expect monotonic increase"
        assert msg["task_id"] == fut.parsl_executor_task_id
        assert msg["context"] == ctxt, "Expect no modification"
        assert msg["buffer"] == buf, "Expect no modification"



from __future__ import annotations

import datetime
import logging
import threading
import time
from abc import abstractmethod, abstractproperty
from concurrent.futures import Future
from functools import cached_property
from itertools import compress
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

from parsl.executors.base import ParslExecutor
from parsl.executors.errors import BadStateException, ScalingFailed
from parsl.jobs.error_handlers import noop_error_handler, simple_error_handler
from parsl.jobs.states import TERMINAL_STATES, JobState, JobStatus
from parsl.monitoring.message_type import MessageType
from parsl.monitoring.radios.multiprocessing import MultiprocessingQueueRadioSender
from parsl.providers.base import ExecutionProvider
from parsl.utils import AtomicIDCounter

logger = logging.getLogger(__name__)


class BlockProviderExecutor(ParslExecutor):
    """A base class for executors which scale using blocks.

    This base class is intended to help with executors which:

    - use blocks of workers to execute tasks
    - blocks of workers are launched on a batch system through
      an `ExecutionProvider`

    An implementing class should implement the abstract methods required by
    `ParslExecutor` to submit tasks, as well as BlockProviderExecutor
    abstract methods to provide the executor-specific command to start a block
    of workers (the ``_get_launch_command`` method), and some basic scaling
    information (``outstanding`` and ``workers_per_node`` properties).

    This base class provides a ``scale_out`` method which will launch new
    blocks. It does not provide a ``scale_in`` method, because scale-in
    behaviour is not well defined in the Parsl scaling model and so behaviour
    is left to individual executors.

    Parsl scaling will provide scaling between min_blocks and max_blocks by
    invoking scale_out, but it will not initialize the blocks requested by
    any init_blocks parameter. Subclasses must implement that behaviour
    themselves.
    """

    def __init__(self, *,
                 provider: Optional[ExecutionProvider],
                 block_error_handler: Union[bool, Callable[[BlockProviderExecutor, Dict[str, JobStatus]], None]]):
        super().__init__()
        self._provider = provider
        self.block_error_handler: Callable[[BlockProviderExecutor, Dict[str, JobStatus]], None]
        if isinstance(block_error_handler, bool):
            if block_error_handler:
                self.block_error_handler = simple_error_handler
            else:
                self.block_error_handler = noop_error_handler
        else:
            self.block_error_handler = block_error_handler

        self._executor_bad_state = threading.Event()
        self._executor_exception: Optional[Exception] = None

        self._block_id_counter = AtomicIDCounter()

        self._tasks = {}  # type: Dict[object, Future]

        self._last_poll_time = 0.0

        # these four structures track, in loosely coordinated fashion, the
        # existence of blocks and jobs and how to map between their
        # identifiers.
        self.blocks_to_job_id = {}  # type: Dict[str, str]
        self.job_ids_to_block = {}  # type: Dict[str, str]

        # errors can happen during the submit call to the provider; this is used
        # to keep track of such errors so that they can be handled in one place
        # together with errors reported by status()
        self._simulated_status: Dict[str, JobStatus] = {}

        # this stores an approximation (sometimes delayed) of the latest status
        # of pending, active and recently terminated blocks
        self._status = {}  # type: Dict[str, JobStatus]

        self.submit_monitoring_radio: Optional[MultiprocessingQueueRadioSender] = None

    @cached_property
    def _scale_prefix(self) -> str:
        """Prefix used for logging messages about scaling"""
        return f'[Scaling executor {self.label}]'

    def start(self):
        super().start()
        if self.monitoring_messages:
            self.submit_monitoring_radio = MultiprocessingQueueRadioSender(self.monitoring_messages)

    def _make_status_dict(self, block_ids: Sequence[str], status_list: Sequence[JobStatus]) -> Dict[str, JobStatus]:
        """Given a sequence of block ids and a sequence of corresponding status strings,
        returns a dictionary mapping each block id to the corresponding status

        :param block_ids: the sequence of block ids
        :param status_list: the sequence of job status strings
        :return: the resulting dictionary
        """
        if len(block_ids) != len(status_list):
            raise IndexError("block id list and status string list differ in size")
        d = {}
        for i in range(len(block_ids)):
            d[block_ids[i]] = status_list[i]

        return d

    @property
    def status_polling_interval(self):
        """Returns the interval, in seconds, at which the status method should be called. The
        assumption here is that, once initialized, an executor's polling interval is fixed.
        In practice, at least given the current situation, the executor uses a single task provider
        and this method is a delegate to the corresponding method in the provider.

        :return: the number of seconds to wait between calls to status() or zero if no polling
                 should be done
        """
        if self._provider is None:
            return 0
        else:
            return self._provider.status_polling_interval

    @abstractmethod
    def outstanding(self) -> int:
        """This should return the number of tasks that the executor has been given to run (waiting to run, and running now)"""

        raise NotImplementedError("Classes inheriting from BlockProviderExecutor must implement "
                                  "outstanding()")

    def set_bad_state_and_fail_all(self, exception: Exception):
        """Allows external error handlers to mark this executor as irrecoverably bad and cause
        all tasks submitted to it now and in the future to fail. The executor is responsible
        for checking  :method:bad_state_is_set() in the :method:submit() method and raising the
        appropriate exception, which is available through :method:executor_exception().
        """
        logger.exception("Setting bad state due to exception", exc_info=exception)
        self._executor_exception = exception
        # Set bad state to prevent new tasks from being submitted
        self._executor_bad_state.set()
        # We set all current tasks to this exception to make sure that
        # this is raised in the main context.
        for task in self._tasks:
            self._tasks[task].set_exception(BadStateException(self, self._executor_exception))

    @property
    def bad_state_is_set(self):
        """Returns true if this executor is in an irrecoverable error state. If this method
        returns true, :property:executor_exception should contain an exception indicating the
        cause.
        """
        return self._executor_bad_state.is_set()

    @property
    def executor_exception(self):
        """Returns an exception that indicates why this executor is in an irrecoverable state."""
        return self._executor_exception

    def handle_errors(self, status: Dict[str, JobStatus]) -> None:
        """This method is called by the error management infrastructure after a status poll. The
        executor implementing this method is then responsible for detecting abnormal conditions
        based on the status of submitted jobs. If the executor does not implement any special
        error handling, this method should return False, in which case a generic error handling
        scheme will be used.
        :param status: status of all jobs launched by this executor
        """
        self.block_error_handler(self, status)

    @property
    def tasks(self) -> Dict[object, Future]:
        return self._tasks

    @property
    def provider(self) -> Optional[ExecutionProvider]:
        return self._provider

    def _filter_scale_in_ids(self, to_kill: Sequence[Any], killed: Sequence[bool]) -> Sequence[Any]:
        """ Filter out job id's that were not killed
        """
        assert len(to_kill) == len(killed)

        if False in killed:
            killed_job_ids = [jid for jid, k in zip(to_kill, killed) if k]
            not_killed_job_ids = [jid for jid, k in zip(to_kill, killed) if not k]
            logger.warning("Some jobs were not killed successfully: "
                           f"killed jobs: {killed_job_ids}, "
                           f"not-killed jobs: {not_killed_job_ids}")

        # Filters first iterable by bool values in second
        return list(compress(to_kill, killed))

    def scale_out_facade(self, n: int) -> List[str]:
        """Scales out the number of blocks by "blocks"
        """
        assert self.provider is not None, \
            "scale_out_facade should only be called on executors with a valid provider"
        block_ids = []
        monitoring_status_changes = {}
        logger.info("%s Scaling out by %d blocks", self._scale_prefix, n)
        for _ in range(n):
            block_id = str(self._block_id_counter.get_id())
            logger.info("%s Allocated block ID %s", self._scale_prefix, block_id)
            try:
                job_id = self._launch_block(block_id)

                pending_status = JobStatus(JobState.PENDING)

                self.blocks_to_job_id[block_id] = job_id
                self.job_ids_to_block[job_id] = block_id
                self._status[block_id] = pending_status

                monitoring_status_changes[block_id] = pending_status
                block_ids.append(block_id)

            except Exception as ex:
                failed_status = JobStatus(JobState.FAILED, "Failed to start block {}: {}".format(block_id, ex))
                self._simulated_status[block_id] = failed_status
                self._status[block_id] = failed_status

        self.send_monitoring_info(monitoring_status_changes)
        return block_ids

    def scale_in(self, blocks: int) -> List[str]:
        """Scale in method.

        Cause the executor to reduce the number of blocks by count.

        The default implementation will kill blocks without regard to their
        status or whether they are executing tasks. Executors with more
        nuanced scaling strategies might overload this method to work with
        that strategy - see the HighThroughputExecutor for an example of that.

        :return: A list of block ids corresponding to the blocks that were removed.
        """

        active_blocks = [block_id for block_id, status in self._status.items()
                         if status.state not in TERMINAL_STATES]

        block_ids_to_kill = active_blocks[:blocks]

        job_ids_to_kill = [self.blocks_to_job_id[block] for block in block_ids_to_kill]

        # Cancel the blocks provisioned
        if self.provider:
            logger.info("%s Scaling in jobs: %s", self._scale_prefix, str(job_ids_to_kill))
            r = self.provider.cancel(job_ids_to_kill)
            job_ids = self._filter_scale_in_ids(job_ids_to_kill, r)
            block_ids_killed = [self.job_ids_to_block[job_id] for job_id in job_ids]
            return block_ids_killed
        else:
            logger.error("%s No execution provider available to scale in", self._scale_prefix)
            return []

    def _launch_block(self, block_id: str) -> Any:
        assert self.provider is not None, \
            "_launch_block should only be called on executors with a valid provider"

        launch_cmd = self._get_launch_command(block_id)
        job_name = f"parsl.{self.label}.block-{block_id}"
        logger.debug("%s Submitting to provider with job_name %s", self._scale_prefix, job_name)
        job_id = self.provider.submit(launch_cmd, 1, job_name)
        if job_id:
            logger.debug("%s Launched block %s with job ID %s", self._scale_prefix, block_id, str(job_id))
        else:
            raise ScalingFailed(self, "Attempt to provision nodes did not return a job ID")
        return job_id

    @abstractmethod
    def _get_launch_command(self, block_id: str) -> str:
        pass

    def _get_block_and_job_ids(self) -> Tuple[List[str], List[Any]]:
        # Not using self.blocks.keys() and self.blocks.values() simultaneously
        # The dictionary may be changed during invoking this function
        # As scale_in and scale_out are invoked in multiple threads
        block_ids = list(self.blocks_to_job_id.keys())
        job_ids = []  # types: List[Any]
        for bid in block_ids:
            job_ids.append(self.blocks_to_job_id[bid])
        return block_ids, job_ids

    @abstractproperty
    def workers_per_node(self) -> Union[int, float]:
        pass

    def send_monitoring_info(self, status: Dict) -> None:
        # Send monitoring info for HTEX when monitoring enabled
        if self.submit_monitoring_radio:
            msg = self.create_monitoring_info(status)
            logger.debug("Sending block monitoring message: %r", msg)
            self.submit_monitoring_radio.send((MessageType.BLOCK_INFO, msg))

    def create_monitoring_info(self, status: Dict[str, JobStatus]) -> Sequence[Dict[str, Any]]:
        """Create a monitoring message for each block based on the poll status.
        """
        return [
            {
                "run_id": self.run_id,
                "status": s.status_name,
                "timestamp": datetime.datetime.now(),
                "executor_label": self.label,
                "job_id": self.blocks_to_job_id.get(bid, None),
                "block_id": bid
            }
            for bid, s in status.items()
        ]

    def poll_facade(self) -> None:
        now = time.time()
        if now >= self._last_poll_time + self.status_polling_interval:
            previous_status = self._status
            self._status = self.status()
            self._last_poll_time = now
            delta_status = {}
            for block_id in self._status:
                if block_id not in previous_status \
                        or previous_status[block_id].state != self._status[block_id].state:
                    delta_status[block_id] = self._status[block_id]

            if delta_status:
                self.send_monitoring_info(delta_status)

    def status(self) -> Dict[str, JobStatus]:
        """Return the status of all jobs/blocks currently known to this executor.

        :return: a dictionary mapping block ids (in string) to job status
        """
        if self._provider:
            block_ids, job_ids = self._get_block_and_job_ids()
            status = self._make_status_dict(block_ids, self._provider.status(job_ids))
        else:
            status = {}
        status.update(self._simulated_status)

        return status

    @property
    def status_facade(self) -> Dict[str, JobStatus]:
        """Return the status of all jobs/blocks of the executor of this poller.

        :return: a dictionary mapping block ids (in string) to job status
        """
        return self._status

    def scale_in_facade(self, n: int, max_idletime: Optional[float] = None) -> List[str]:

        if max_idletime is None:
            block_ids = self.scale_in(n)
        else:
            # This is a HighThroughputExecutor-specific interface violation.
            # This code hopes, through pan-codebase reasoning, that this
            # scale_in method really does come from HighThroughputExecutor,
            # and so does have an extra max_idletime parameter not present
            # in the executor interface.
            block_ids = self.scale_in(n, max_idletime=max_idletime)  # type: ignore[call-arg]
        if block_ids is not None:
            new_status = {}
            for block_id in block_ids:
                logger.debug("Marking block %s as SCALED_IN", block_id)
                s = JobStatus(JobState.SCALED_IN)
                new_status[block_id] = s
                self._status[block_id] = s
                self._simulated_status[block_id] = s
            self.send_monitoring_info(new_status)
        return block_ids



import random

import pytest

import parsl
from parsl.config import Config
from parsl.executors.high_throughput.executor import HighThroughputExecutor
from parsl.providers import LocalProvider


@pytest.mark.local
def test_max_workers_per_node():

    w = random.randint(1, 8)
    h = HighThroughputExecutor(max_workers_per_node=w,
                               provider=LocalProvider(max_blocks=0, init_blocks=0, min_blocks=0)
                               )
    c = Config(executors=[h])

    with parsl.load(c):
        assert h.max_workers_per_node == w, "max_workers_per_node should be wired all the way through"
        assert h.workers_per_node == w, "max_workers_per_node is chosen in the absence of other restrictions"


@pytest.mark.local
def test_cores():

    p = LocalProvider(max_blocks=0, init_blocks=0, min_blocks=0)
    p.cores_per_node = 8

    h = HighThroughputExecutor(max_workers_per_node=16,
                               provider=p,
                               cores_per_worker=4
                               )
    c = Config(executors=[h])

    with parsl.load(c):
        assert h.cores_per_worker == 4, "cores_per_worker should be wired all the way through"
        assert h.workers_per_node == 2, "8 cores per node, 4 cores per worker => 2 workers"


@pytest.mark.local
def test_mem():

    p = LocalProvider(max_blocks=0, init_blocks=0, min_blocks=0)
    p.cores_per_node = 8
    p.mem_per_node = 16

    h = HighThroughputExecutor(max_workers_per_node=16,
                               provider=p,
                               cores_per_worker=4,
                               mem_per_worker=16
                               )
    c = Config(executors=[h])

    with parsl.load(c):
        assert h.mem_per_worker == 16, "mem_per_worker should be wired all the way through"
        assert h.workers_per_node == 1, "one worker gets all the memory"


@pytest.mark.local
def test_accelerators():

    p = LocalProvider(max_blocks=0, init_blocks=0, min_blocks=0)
    p.cores_per_node = 16
    p.mem_per_node = 16

    h = HighThroughputExecutor(max_workers_per_node=16,
                               provider=p,
                               cores_per_worker=4,
                               mem_per_worker=2,
                               available_accelerators=3
                               )
    c = Config(executors=[h])

    with parsl.load(c):
        assert h.workers_per_node == 3, "number of accelerators is worker constraining limit here"



import time

import pytest

import parsl
from parsl.config import Config
from parsl.executors import HighThroughputExecutor
from parsl.launchers import WrappedLauncher
from parsl.providers import LocalProvider


def local_config():
    # see the comments inside test_regression for reasoning about why each
    # of these parameters is set why it is.
    return Config(
        max_idletime=1,

        strategy='htex_auto_scale',
        strategy_period=1,

        executors=[
            HighThroughputExecutor(
                label="htex_local",
                encrypted=True,
                provider=LocalProvider(
                    init_blocks=1,
                    min_blocks=0,
                    max_blocks=1,
                    launcher=WrappedLauncher(prepend="sleep inf ; "),
                ),
            )
        ],
    )


@parsl.python_app
def task():
    return 7


@pytest.mark.local
def test_regression(try_assert):
    # The above config means that we should start scaling out one initial
    # block, but then scale it back in after a second or so if the executor
    # is kept idle (which this test does using try_assert).

    # Because of 'sleep inf' in the WrappedLaucher, the block will not ever
    # register.

    # The bug being tested is about mistreatment of blocks which are scaled in
    # before they have a chance to register, and the above forces that to
    # happen.

    # After that scaling in has happened, we should see that we have one block
    # and it should be in a terminal state. The below try_assert waits for
    # that to become true.

    # At that time, we should also see htex reporting no blocks registered - as
    # mentioned above, that is a necessary part of the bug being tested here.

    # Give 10 strategy periods for the above to happen: each step of scale up,
    # and scale down due to idleness isn't guaranteed to happen in exactly one
    # scaling step.

    htex = parsl.dfk().executors['htex_local']

    try_assert(lambda: len(htex.status_facade) == 1 and htex.status_facade['0'].terminal,
               timeout_ms=10000)

    assert htex.connected_blocks() == [], "No block should have connected to interchange"

    # Now we can reconfigure the launcher to let subsequent blocks launch ok,
    # and run a trivial task. That trivial task will scale up a new block and
    # run the task successfully.

    # Prior to issue #3568, the bug was that the scale in of the first
    # block earlier in the test case would have incorrectly been treated as a
    # failure, and then the block error handler would have treated that failure
    # as a permanent htex failure, and so the task execution below would raise
    # a BadStateException rather than attempt to run the task.

    assert htex.provider.launcher.prepend != "", "Pre-req: prepend attribute should exist and be non-empty"
    htex.provider.launcher.prepend = ""
    assert task().result() == 7



import argparse

import pytest

import parsl
from parsl.app.app import python_app
from parsl.tests.configs.local_threads import config


@python_app
def increment(x):
    return x + 1


@python_app
def slow_increment(x):
    import time
    time.sleep(0.001)
    return x + 1


@pytest.mark.parametrize("depth", (2, 3))
def test_increment(depth):
    """Test simple pipeline A->B...->N"""
    futs = [increment(0)]
    futs.extend(increment(futs[i - 1]) for i in range(1, depth))
    assert sum(f.result() for f in futs) == sum(range(1, depth + 1))


@pytest.mark.parametrize("depth", (2, 3))
def test_increment_slow(depth):
    """Test simple pipeline A->B...->N with delay"""
    futs = [slow_increment(0)]
    futs.extend(slow_increment(futs[i - 1]) for i in range(1, depth))
    assert sum(f.result() for f in futs) == sum(range(1, depth + 1))



import pytest

from parsl.app.app import bash_app
from parsl.app.futures import DataFuture
from parsl.data_provider.files import File


@bash_app
def increment(inputs=(), outputs=(), stdout=None, stderr=None):
    cmd_line = """
    if ! [ -f {inputs[0]} ] ; then exit 43 ; fi
    x=$(cat {inputs[0]})
    echo $(($x+1)) > {outputs[0]}
    """.format(inputs=inputs, outputs=outputs)
    return cmd_line


@bash_app
def slow_increment(dur, inputs=(), outputs=(), stdout=None, stderr=None):
    cmd_line = """
    x=$(cat {inputs[0]})
    echo $(($x+1)) > {outputs[0]}
    sleep {0}
    """.format(dur, inputs=inputs, outputs=outputs)
    return cmd_line


@pytest.mark.staging_required
def test_increment(tmpd_cwd, depth=5):
    """Test simple pipeline A->B...->N
    """
    fpath = tmpd_cwd / "test0.txt"
    fpath.write_text("0\n")

    prev = [File(str(fpath))]
    futs = []
    for i in range(1, depth):
        assert isinstance(prev[0], (DataFuture, File))
        output = File(str(tmpd_cwd / f"test{i}.txt"))
        f = increment(
            inputs=prev,
            outputs=[output],
            stdout=str(tmpd_cwd / f"incr{i}.out"),
            stderr=str(tmpd_cwd / f"incr{i}.err"),
        )
        prev = f.outputs
        futs.append((i, prev[0]))
        assert isinstance(prev[0], DataFuture)

    for key, f in futs:
        file = f.result()
        expected = str(tmpd_cwd / f"test{key}.txt")

        assert file.local_path is None, "File on local side has overridden local_path, file: {}".format(repr(file))
        assert file.filepath == expected, "Submit side filepath has not been preserved over execution"
        data = open(file.filepath).read().strip()
        assert data == str(key)


@pytest.mark.staging_required
def test_increment_slow(tmpd_cwd, depth=5, dur=0.01):
    """Test simple pipeline slow (sleep.5) A->B...->N
    """

    fpath = tmpd_cwd / "test0.txt"
    fpath.write_text("0\n")

    prev = [File(str(fpath))]
    futs = []
    for i in range(1, depth):
        output = File(str(tmpd_cwd / f"test{i}.txt"))
        f = slow_increment(
            dur,
            inputs=prev,
            outputs=[output],
            stdout=str(tmpd_cwd / f"incr{i}.out"),
            stderr=str(tmpd_cwd / f"incr{i}.err"),
        )
        prev = f.outputs
        futs.append((i, prev[0]))

    for key, f in futs:
        data = open(f.result().filepath).read().strip()
        assert data == str(key)



import contextlib
import os

import pytest

import parsl
from parsl import python_app
from parsl.config import Config
from parsl.dataflow.memoization import BasicMemoizer
from parsl.executors.threads import ThreadPoolExecutor
from parsl.utils import get_last_checkpoint


def parsl_configured(run_dir, memoizer):
    return parsl.load(Config(
        run_dir=str(run_dir),
        executors=[
            ThreadPoolExecutor(
                label='local_threads_checkpoint',
            )
        ],
        memoizer=memoizer
    ))


@python_app(cache=True)
def uuid_app(x):
    import uuid
    return uuid.uuid4()


@pytest.mark.local
def test_loading_checkpoint(tmpd_cwd):
    """Tests that get_last_checkpoint finds checkpoints from prior run
    """
    import uuid
    param = str(uuid.uuid4())

    with parsl_configured(tmpd_cwd, BasicMemoizer(checkpoint_mode="task_exit", checkpoint_files=[])):
        result = uuid_app(param).result()

    with parsl_configured(tmpd_cwd, BasicMemoizer(checkpoint_mode="task_exit", checkpoint_files=get_last_checkpoint(rundir=str(tmpd_cwd)))):
        relaunched = uuid_app(param).result()

    assert result == relaunched, "Expected following call to uuid_app to return cached uuid"

    # check that we don't lose the result just because it came from two runs ago,
    # by running the same test again.
    with parsl_configured(tmpd_cwd, BasicMemoizer(checkpoint_mode="task_exit", checkpoint_files=get_last_checkpoint(rundir=str(tmpd_cwd)))):
        relaunched = uuid_app(param).result()

    assert result == relaunched, "Expected following call to uuid_app to return cached uuid"


@pytest.mark.local
def test_loading_checkpoint_other_files(tmpd_cwd):
    """Tests that get_last_checkpoint ignores other files.

       This is a regression test for a bug where the alphabetically last
       per-run directory was being selected, and then ignored if it was
       not a checkpoint directory, for example when the monitoring.db
       file is also present (and alphabetically after any numbered
       run directories).
    """
    import uuid
    param = str(uuid.uuid4())

    with parsl_configured(tmpd_cwd, BasicMemoizer(checkpoint_mode="task_exit", checkpoint_files=[])):
        result = uuid_app(param).result()

    with open(tmpd_cwd / "X", "w"):
        # don't need to write anything, as we only need the directory entry
        # which is made by the open call.
        pass

    with parsl_configured(tmpd_cwd, BasicMemoizer(checkpoint_mode="task_exit", checkpoint_files=get_last_checkpoint(rundir=str(tmpd_cwd)))):
        relaunched = uuid_app(param).result()

    assert result == relaunched, "Expected following call to uuid_app to return cached uuid"


@pytest.mark.local
def test_high_rundir(tmpd_cwd):
    """Tests that get_last_checkpoint respects rundirs of more than 3 digits.
    """
    import uuid
    param = str(uuid.uuid4())

    with open(tmpd_cwd / "998", "w"):
        pass

    # this will have rundir "999"
    with parsl_configured(tmpd_cwd, BasicMemoizer(checkpoint_mode="task_exit", checkpoint_files=[])):
        assert "999" in parsl.dfk().run_dir
        result_999 = uuid_app(param).result()

    # this will have rundir "1000", lexically before 999
    with parsl_configured(tmpd_cwd, BasicMemoizer(checkpoint_mode="task_exit", checkpoint_files=[])):
        assert "1000" in parsl.dfk().run_dir
        result_1000 = uuid_app(param).result()

    assert result_999 != result_1000, "Should have not used previous result"

    with parsl_configured(tmpd_cwd, BasicMemoizer(checkpoint_mode="task_exit", checkpoint_files=get_last_checkpoint(rundir=str(tmpd_cwd)))):
        relaunched = uuid_app(param).result()

    assert result_1000 == relaunched, "Expected restore to be from 1000 rundir"



from argparse import ArgumentParser

import parsl

from library.config import make_local_config
from library.app import convert_many_to_binary
from parsl.app.python import PythonApp

# Protect the script from running twice.
#  See "Safe importing of main module" in Python multiprocessing docs
#  https://docs.python.org/3/library/multiprocessing.html#multiprocessing-programming
if __name__ == "__main__":
    # Get user instructions
    parser = ArgumentParser()
    parser.add_argument('--numbers-per-batch', default=8, type=int)
    parser.add_argument('numbers', nargs='+', type=int)
    args = parser.parse_args()

    # Prepare the workflow functions
    convert_app = PythonApp(convert_many_to_binary, cache=False)

    # Load the configuration
    #  As a context manager so resources are shutdown on exit
    with parsl.load(make_local_config()):

        # Spawn tasks
        futures = [
            convert_app(args.numbers[start:start + args.numbers_per_batch])
            for start in range(0, len(args.numbers), args.numbers_per_batch)
        ]

        # Retrieve task results
        for future in futures:
            for x, b in zip(future.task_record['args'][0], future.result()):
                print(f'{x} -> {"".join("1" if i else "0" for i in b)}')



import os

from parsl.config import Config
from parsl.executors.radical import RadicalPilotExecutor, ResourceConfig

rpex_cfg = ResourceConfig()


def fresh_config():

    return Config(
            executors=[
                RadicalPilotExecutor(
                    label='RPEXBulk',
                    rpex_cfg=rpex_cfg,
                    resource='local.localhost',
                    runtime=30, cores=4)])



import pytest

import parsl
from parsl.app.app import python_app
from parsl.config import Config
from parsl.dataflow.memoization import BasicMemoizer
from parsl.executors.threads import ThreadPoolExecutor


def fresh_config(*, config_run_dir):
    tpe = ThreadPoolExecutor(label='local_threads_checkpoint_periodic', max_threads=1)
    return Config(
        run_dir=config_run_dir,
        executors=[tpe],
        memoizer=BasicMemoizer(
            checkpoint_mode='periodic',
            checkpoint_period='00:00:02'
        )
    )


@python_app(cache=True)
def slow_double(x, sleep_dur=1):
    import time
    time.sleep(sleep_dur)
    return x * 2


def tstamp_to_seconds(line):
    f = line.partition(" ")[0]
    return float(f)


@pytest.mark.local
def test_periodic(tmpd_cwd):
    """Test checkpointing with task_periodic behavior
    """
    with parsl.load(fresh_config(config_run_dir=str(tmpd_cwd))):
        memoizer = parsl.dfk().memoizer
        h, m, s = map(int, memoizer.checkpoint_period.split(":"))
        assert h == 0, "Verify test setup"
        assert m == 0, "Verify test setup"
        assert s > 0, "Verify test setup"
        sleep_for = s + 1
        futs = [slow_double(sleep_for) for _ in range(4)]
        [f.result() for f in futs]
        run_dir = parsl.dfk().run_dir

    # Here we will check if the loglines came back with 5 seconds deltas
    with open("{}/parsl.log".format(run_dir)) as f:
        log_lines = f.readlines()
    expected_msg = " Done checkpointing"
    expected_msg2 = " No tasks checkpointed in this pass"

    lines = [line for line in log_lines if expected_msg in line or expected_msg2 in line]
    assert len(lines) >= 3, "Insufficient checkpoint lines in logfile"
    deltas = [tstamp_to_seconds(line) for line in lines]
    assert deltas[1] - deltas[0] < 5.5, "Delta between checkpoints exceeded period"
    assert deltas[2] - deltas[1] < 5.5, "Delta between checkpoints exceeded period"



import pathlib

import pytest

import parsl
from parsl.config import Config
from parsl.executors import ThreadPoolExecutor
from parsl.logconfigs.json import JSONLogging
from parsl.monitoring import MonitoringHub


def config(*, run_dir):
    c = Config(executors=[ThreadPoolExecutor()],
               monitoring=MonitoringHub(),
               run_dir=str(run_dir))
    return c


def config_log(*, run_dir, log_config):
    c = Config(executors=[ThreadPoolExecutor()],
               monitoring=MonitoringHub(),
               run_dir=str(run_dir),
               initialize_logging=log_config)
    return c


@pytest.mark.local
def test_log_default_config(tmpd_cwd, try_assert):
    with parsl.load(config(run_dir=tmpd_cwd)):
        try_assert(lambda: (pathlib.Path(parsl.dfk().run_dir) / "database_manager.log").exists())


@pytest.mark.local
def test_log_json_config(tmpd_cwd, try_assert):
    with parsl.load(config_log(run_dir=tmpd_cwd, log_config=JSONLogging())):
        try_assert(lambda: (pathlib.Path(parsl.dfk().run_dir) / "database_manager.jsonlog").exists())



import logging
from typing import List, Optional, Sequence, Union

from parsl.executors.status_handling import BlockProviderExecutor
from parsl.jobs.strategy import Strategy
from parsl.utils import Timer

logger = logging.getLogger(__name__)


class JobStatusPoller(Timer):

    def __init__(self, *, strategy: Optional[str], max_idletime: float,
                 strategy_period: Union[float, int]) -> None:
        """Initalize the poller.

        strategy: str
          names the scaling strategy to use

        max_idletime: float, seconds
          used by some scaling strategies to decide how long a block is idle

        strategy_period: float|int, seconds
          how often the scaling strategy should be executed
        """
        # only executors which should be polled and which have a valid
        # provider should be added to this list: so it is safe to assume
        # that e.provider is not None for e in self._executors
        self._executors: List[BlockProviderExecutor] = []

        self._strategy = Strategy(strategy=strategy,
                                  max_idletime=max_idletime)
        super().__init__(self.poll, interval=strategy_period, name="JobStatusPoller")

    def poll(self) -> None:
        for executor in self._executors:
            # Update status
            executor.poll_facade()
            executor.handle_errors(executor.status_facade)

        self._strategy.strategize(self._executors)

    def add_executors(self, executors: Sequence[BlockProviderExecutor]) -> None:
        for executor in executors:
            if executor.status_polling_interval > 0:
                assert executor.provider is not None, \
                    "BlockProviderExecutor.status_polling_interval implementation guarantee: None => status_polling_interval == 0"
                logger.debug("Adding executor {}".format(executor.label))
                self._executors.append(executor)
        self._strategy.add_executors(executors)

    def close(self, timeout: Optional[float] = None) -> None:
        super().close(timeout)
        for executor in self._executors:
            if not executor.bad_state_is_set:
                logger.info(f"Scaling in executor {executor.label}")

                # this code needs to be at least as many blocks as need
                # cancelling, but it is safe to be more, as the scaling
                # code will cope with being asked to cancel more blocks
                # than exist.
                block_count = len(executor.status_facade)
                executor.scale_in_facade(block_count)

            else:  # and bad_state_is_set
                logger.warning(f"Not scaling in executor {executor.label} because it is in bad state")



import pytest

from parsl.monitoring.message_type import MessageType
from parsl.monitoring.radios.filesystem import FilesystemRadio
from parsl.multiprocessing import SpawnQueue


@pytest.mark.local
def test_filesystem(tmpd_cwd):
    """Test filesystem radio/receiver pair.
    This test checks that the pair can be started up locally, that a message
    is conveyed from radio to receiver, and that the receiver process goes
    away at shutdown.
    """

    resource_msgs = SpawnQueue()

    radio_config = FilesystemRadio()

    # start receiver
    receiver = radio_config.create_receiver(run_dir=str(tmpd_cwd),
                                            resource_msgs=resource_msgs)

    # make radio

    radio_sender = radio_config.create_sender()

    # send message into radio

    message = (MessageType.RESOURCE_INFO, {})

    radio_sender.send(message)

    # verify it comes out of the receiver

    m = resource_msgs.get()

    assert m == message, "The sent message should appear in the queue"

    # shut down router

    receiver.shutdown()

    # we can't inspect the process if it has been closed properly, but
    # we can verify that it raises the expected ValueError the closed
    # processes raise on access.
    with pytest.raises(ValueError):
        receiver.process.exitcode



import os
import pickle
import sys
from argparse import ArgumentError
from typing import Any, Dict
from unittest import mock

import pytest

from parsl.app.errors import RemoteExceptionWrapper
from parsl.executors.high_throughput import process_worker_pool
from parsl.executors.high_throughput.process_worker_pool import worker
from parsl.multiprocessing import SpawnContext
from parsl.serialize.facade import deserialize

if sys.version_info < (3, 12):
    # exit_on_error bug; see https://github.com/python/cpython/issues/121018
    # "argparse.ArgumentParser.parses_args does not honor exit_on_error=False when
    #  given unrecognized arguments"
    pytest.skip(allow_module_level=True, reason="exit_on_error argparse bug")

# due to above pytest.skip, mypy on < Py312 thinks this is unreachable.  :facepalm:
_known_required = (  # type: ignore[unreachable, unused-ignore]
    "--cert_dir",
    "--cpu-affinity",
    "--port",
    "--addresses",
)


@pytest.mark.local
def test_arg_parser_exits_on_error():
    p = process_worker_pool.get_arg_parser()
    assert p.exit_on_error


@pytest.mark.local
def test_arg_parser_known_required():
    p = process_worker_pool.get_arg_parser()
    reqd = [a for a in p._actions if a.required]
    for a in reqd:
        assert a.option_strings[-1] in _known_required, "Update _known_required?"


@pytest.mark.local
@pytest.mark.parametrize("req", _known_required)
def test_arg_parser_required(req):
    p = process_worker_pool.get_arg_parser()
    p.exit_on_error = False
    with pytest.raises(ArgumentError) as pyt_exc:
        p.parse_args([])

    e_msg = pyt_exc.value.args[1]
    assert req in e_msg


@pytest.mark.local
@pytest.mark.parametrize("valid,val", (
        (True, "NoNe"),
        (True, "none"),
        (True, "block"),
        (True, "alternating"),
        (True, "block-reverse"),
        (True, "list"),
        (False, "asdf"),
        (False, ""),
))
def test_arg_parser_validates_cpu_affinity(valid, val):
    reqd_args = []
    reqd_args.extend(("--cert_dir", "/some/path"))
    reqd_args.extend(("--port", "123"))
    reqd_args.extend(("--addresses", "asdf"))
    reqd_args.extend(("--cpu-affinity", val))

    p = process_worker_pool.get_arg_parser()
    p.exit_on_error = False
    if valid:
        p.parse_args(reqd_args)
    else:
        with pytest.raises(ArgumentError) as pyt_exc:
            p.parse_args(reqd_args)
        assert "must be one of" in pyt_exc.value.args[1]


def _always_raise(*a, **k):
    raise ArithmeticError(f"{a=}\n{k=}")


@pytest.mark.local
def test_worker_dynamic_import_happy_path(tmpd_cwd) -> None:
    import_str = f"{_always_raise.__module__}.{_always_raise.__name__}"
    task_exec: Dict[str, Any] = {
        "f": import_str,
        "a": (1, 2),
        "k": {"a": "b"},
    }
    req = {
        "task_id": 15,
        "context": {"task_executor": task_exec},
        "buffer": b"some serialized value"
    }

    try:
        task_args = [req["buffer"]]
        task_args.extend(task_exec["a"])
        _always_raise(*task_args, **task_exec["k"])
    except Exception as e:
        exp_exc = e
    else:
        raise RuntimeError("Test failure; this branch should not run")

    q = mock.Mock(side_effect=(req, MemoryError("intentional test error")))
    q.get = q

    block_id = "bid"
    worker_id = 1
    pool = 1
    (tmpd_cwd / f"block-{block_id}/{worker_id}").mkdir(parents=True)
    with pytest.raises(MemoryError):
        worker(
            worker_id,
            pool_id=str(pool),
            pool_size=pool,
            task_queue=q,
            result_queue=q,
            monitoring_queue=None,
            ready_worker_count=SpawnContext.Value("i", 0),
            tasks_in_progress={},
            cpu_affinity="none",
            accelerator=None,
            block_id=block_id,
            task_queue_timeout=0,
            manager_pid=os.getpid(),
            logdir=str(tmpd_cwd),
            log_config=None,
            debug=True,
            mpi_launcher="",
        )
    (result_pkl,), _ = q.put.call_args
    r = pickle.loads(result_pkl)
    assert "exception" in r
    wrapped_exc: RemoteExceptionWrapper = deserialize(r["exception"])
    exc = wrapped_exc.get_exception()
    assert isinstance(exc, type(exp_exc)), "Approximate equality"
    assert str(exp_exc) == str(exc), "Approximate equality; all args, kwargs conveyed"


@pytest.mark.local
def test_worker_bad_dynamic_import(tmpd_cwd) -> None:
    req = {
        "task_id": 15,
        "context": {
            "task_executor": {
                "f": "parsl.some.not_existing.module.__nope",
                "a": (1, 2),
                "k": {"a": "b"},
            },
        },
        "buffer": b"some serialized value"
    }

    q = mock.Mock(side_effect=(req, MemoryError("intentional test error")))
    q.get = q

    block_id = "bid"
    worker_id = 1
    pool = 1
    (tmpd_cwd / f"block-{block_id}/{worker_id}").mkdir(parents=True)
    with pytest.raises(MemoryError):
        worker(
            worker_id,
            pool_id=str(pool),
            pool_size=pool,
            task_queue=q,
            result_queue=q,
            monitoring_queue=None,
            ready_worker_count=SpawnContext.Value("i", 0),
            tasks_in_progress={},
            cpu_affinity="none",
            accelerator=None,
            block_id=block_id,
            task_queue_timeout=0,
            manager_pid=os.getpid(),
            logdir=str(tmpd_cwd),
            log_config=None,
            debug=True,
            mpi_launcher="",
        )
    (result_pkl,), _ = q.put.call_args
    r = pickle.loads(result_pkl)
    assert "exception" in r
    wrapped_exc: RemoteExceptionWrapper = deserialize(r["exception"])
    exc = wrapped_exc.get_exception()
    assert isinstance(exc, ModuleNotFoundError)
    assert "No module named" in str(exc)



"""Tests documentation related to building apps. Must reside outside the Parsl library to be effective"""
from typing import List, Union

import numpy as np

import parsl
from parsl import Config, HighThroughputExecutor, python_app

parsl.load(Config(executors=[HighThroughputExecutor(label='htex_spawn', max_workers_per_node=1, address='127.0.0.1')]))


# Part 1: Explain imports
# BAD: Assumes library has been imported
@python_app(executors=['htex_spawn'])
def bad_imports(x: Union[List[float], np.ndarray], m: float, b: float):
    return np.multiply(x, m) + b


# GOOD: Imports libraries itself
@python_app(executors=['htex_spawn'])
def good_imports(x: Union[List[float], 'np.ndarray'], m: float, b: float):
    import numpy as np
    return np.multiply(x, m) + b


future = bad_imports([1.], 1, 0)

try:
    future.result()
    raise ValueError()
except NameError as e:
    print('Failed, as expected. Error:', e)

future = good_imports([1.], 1, 0)
print(f'Passed, as expected: {future.result()}')

# Part 2: Test other types of globals
# BAD: Uses global variables
global_var = {'a': 0}


@python_app
def bad_global(string: str, character: str = 'a'):
    global_var[character] += string.count(character)  # `global_var` will not be accessible


# GOOD
@python_app
def good_global(string: str, character: str = 'a'):
    return {character: string.count(character)}


try:
    bad_global('parsl').result()
except NameError as e:
    print(f'Failed, as expected: {e}')

for ch, co in good_global('parsl', 'a').result().items():
    global_var[ch] += co


# Part 3: Mutable args

# BAD: Assumes changes to inputs will be communicated
@python_app
def append_to_list(input_list: list, new_val):
    input_list.append(new_val)


mutable_arg = []
append_to_list(mutable_arg, 1).result()
assert mutable_arg == [], 'The list _was_changed'


# GOOD: Changes to inputs are returned
@python_app
def append_to_list(input_list: list, new_val) -> list:
    input_list.append(new_val)
    return input_list


mutable_arg = append_to_list(mutable_arg, 1).result()
assert mutable_arg == [1]



import logging
import os
import time

import pytest

import parsl
from parsl.config import Config
from parsl.executors import HighThroughputExecutor
from parsl.launchers import SimpleLauncher
from parsl.monitoring import MonitoringHub
from parsl.providers import LocalProvider


def fresh_config(run_dir, strategy, db_url):
    return Config(
        run_dir=os.fspath(run_dir),
        executors=[
            HighThroughputExecutor(
                label="htex_local",
                cores_per_worker=1,
                encrypted=True,
                provider=LocalProvider(
                    init_blocks=1,
                    # min and max are set to 0 to ensure that we don't get
                    # a block from ongoing strategy scaling, only from
                    # init_blocks
                    min_blocks=0,
                    max_blocks=0,
                    launcher=SimpleLauncher(),
                ),
            )
        ],
        strategy=strategy,
        strategy_period=0.1,
        monitoring=MonitoringHub(
                        logging_endpoint=db_url
        )
    )


@parsl.python_app
def this_app():
    pass


@pytest.mark.local
@pytest.mark.parametrize("strategy", ('none', 'simple', 'htex_auto_scale'))
def test_row_counts(tmpd_cwd, strategy):
    # this is imported here rather than at module level because
    # it isn't available in a plain parsl install, so this module
    # would otherwise fail to import and break even a basic test
    # run.
    import sqlalchemy
    from sqlalchemy import text

    db_url = f"sqlite:///{tmpd_cwd}/monitoring.db"
    with parsl.load(fresh_config(tmpd_cwd, strategy, db_url)):
        dfk = parsl.dfk()
        run_id = dfk.run_id

        this_app().result()

    engine = sqlalchemy.create_engine(db_url)
    with engine.begin() as connection:

        binds = {"run_id": run_id}

        result = connection.execute(text("SELECT COUNT(DISTINCT block_id) FROM block WHERE run_id = :run_id"), binds)
        (c, ) = result.first()
        assert c == 1, "We should see a single block in this database"

        result = connection.execute(text("SELECT COUNT(*) FROM block WHERE block_id = 0 AND status = 'PENDING' AND run_id = :run_id"), binds)
        (c, ) = result.first()
        assert c == 1, "There should be a single pending status"

        result = connection.execute(text("SELECT COUNT(*) FROM block WHERE block_id = 0 AND status = 'SCALED_IN' AND run_id = :run_id"), binds)
        (c, ) = result.first()
        assert c == 1, "There should be a single cancelled status"
